"""
This utility is designed to generate a Giant Extended Degree GARS grid.

##
## GIANT EXTENDED DEGREE GARS
##


# 60 DEGREE CELLS

GED-GARS divides the surface of the earth into 60-degree by 60-degree cells.
Each cell is identified by a four-character designation. (ex. GD1A)

The first two characters are GD to designate that it is using the
giant extended degree system.

The next character designate a 60-degree wide longitudinal band.
Beginning with the 180-degree meridian and proceeding eastward,
the bands are numbered from 1 to 6, so that 180 E to 120 W is band 1;
120 W to 60 W is band 2; and so on.

The fourth character designates a 60-degree wide latitudinal band.
Beginning at the south pole and proceeding northward, the bands are lettered
from A to C (using only letters ABC) so that 90 S to 30 S is band A;
30 S to 30 N is band B; and 30N to 90N is band C.

# 30 DEGREE CELLS

    0  30   60
 60 ---------
    | 1 | 2 |
 30 ---------
    | 3 | 4 |
 0  --------

The areas are numbered sequentially, from west to east, starting with the
northernmost band. The graphical representation of a 30-degree quadrant with
numbered 30-degree by 30-degree areas resembles a telephone keypad.

Each 30-degree by 30-degree area, or keypad key is identified by a five-character
designation. The first four characters comprise the 30-degree quadrant designation.
The fifth character is the keypad key number. (ex. GD6A3)

"""
import math
import re
from typing import Optional, Tuple

import shapely.geometry

from .garsgrid import GARSGridBase


class GEDGARSGrid(GARSGridBase):
    """
    This object gives the polygon of the ED-GARS grid based on the ED-GARS ID
    or the resolution and lat/lon coords for the extended version using
    degrees.

    """

    LETTERS = "ABC"
    VALID_RESOLUTIONS: Tuple[int, int] = (30, 60)
    RE_PATTERN = re.compile(
        r"^GD(?P<quadrant_60deg>\d{1}[A-C])(?P<quadrant_30deg>[1-4])?$"
    )

    def __init__(self, gars_id: str, max_resolution: Optional[int] = None):
        """
        Parameters
        ----------
        gars_id: str
            The string representing the ED-GARS grid.
        max_resolution: int, optional
            The grid resolution in degrees (30, 60).
            If not provided, it will be inferred from the `gars_id`.

        """
        super().__init__()

        max_resolution = (
            None if max_resolution is None else self.validate_resolution(max_resolution)
        )

        # truncate ED-GARS ID to match resolution
        if max_resolution == 60:
            gars_id = gars_id[:4]

        # validate ED-GARS ID
        gars_match = self.RE_PATTERN.match(gars_id)
        if not gars_match:
            raise ValueError(f'"{gars_id}" is not a valid ED-GARS grid ID.')
        self.gars_id: str = gars_id

        gars_dict = gars_match.groupdict()

        self.quadrant_60deg: str = gars_dict["quadrant_60deg"]
        self.quadrant_30deg: str = gars_dict["quadrant_30deg"]

        lon_num = int(self.quadrant_60deg[0])
        if (lon_num < 1) or (lon_num > 6):
            raise ValueError(
                f'"{gars_id}" is not a valid GED-GARS grid ID. '
                "Longitude numbers can only be between 1-6."
            )

        # determine resolution from ED-GARS ID
        self.resolution: int = 60
        if self.quadrant_30deg:
            self.resolution = 30

        # properties
        self._polygon: Optional[shapely.geometry.Polygon] = None

    def __repr__(self):
        return f"<GED-GARS(gars_id={self}, resolution={self.resolution})>"

    @property
    def utm_epsg(self) -> None:  # type: ignore
        """
        Not valid since it spans UTM Zones
        """
        return None

    @classmethod
    def from_latlon(
        cls, latitude: float, longitude: float, resolution: int
    ) -> "GEDGARSGrid":
        """Load ED-GARS grid from latitude and longitude.

        Parameters
        ----------
        latitude: float
            The latitude of the cell you want to find.
        longitude: float
            The longitude of the cell you want to find.
        resolution: int
            The grid resolution in degrees (30, 60).


        Returns
        -------
        GEDGARSGrid

        """
        resolution = cls.validate_resolution(resolution)

        # convert from (-180, 180) to (0, 360)
        longitude = (longitude + 180) % 360 if longitude != 180 else 360
        # convert from (-90, 90) to (0, 180)
        latitude = (latitude + 90) % 180 if latitude != 90 else 179.9999999999

        lon_idx = int(math.floor(longitude / 60.0))
        lat_idx = int(math.floor(latitude / 60.0))
        # 60 deg quadrant
        quadrant_60deg = str(lon_idx + 1) + cls.LETTERS[lat_idx]

        quadrant_30deg = ""
        # 30 deg quadrant
        if resolution < 60:
            lon_30deg_idx = math.floor((longitude % 60) / 30.0) + 1
            lat_30deg_idx = 2 - math.floor((latitude % 60) / 30.0)
            quadrant_30deg = str(int((lat_30deg_idx - 1) * 2 + lon_30deg_idx))

        gars_id = "".join(["GD", quadrant_60deg, quadrant_30deg])

        return cls(gars_id, max_resolution=resolution)

    @property
    def polygon(self) -> shapely.geometry.Polygon:
        """Generates the GARS bounding polygon.

        Returns
        -------
        :obj:`shapely.geometry.Polygon`

        """
        if self._polygon is not None:
            return self._polygon

        # CALCULATE 60 DEG DEGREES
        # get 60 DEG quadrant info
        longitude: float = ((int(self.quadrant_60deg[0]) - 1) * 60) - 180

        # 60 deg north/south letter, A-C
        lat_60deg_letter = self.quadrant_60deg[1]
        latitude: float = -90.0 + (self.LETTERS.index(lat_60deg_letter) * 60.0)

        # CALCULATE 30 DEG DELTA
        try:
            quad_30deg = int(self.quadrant_30deg)
        except TypeError:
            quad_30deg = None

        if quad_30deg is not None:
            if quad_30deg in (2, 4):
                longitude += 30.0

            if quad_30deg in (1, 2):
                latitude += 30.0

        # upper right coordinate
        ul_latitude = latitude + self.resolution
        ul_longitude = longitude + self.resolution

        self._polygon = shapely.geometry.box(
            longitude, latitude, ul_longitude, ul_latitude
        )

        return self._polygon



"""
This utility is designed to generate an Extended Degree GARS grid.

##
## EXTENDED DEGREE GARS
##


# 6 DEGREE CELLS

ED-GARS divides the surface of the earth into 6-degree by 6-degree cells.
Each cell is identified by a five-character designation. (ex. D06AG)

The first character is a D to designate that it is using the extended degree system.

The next two characters designate a 6-degree wide longitudinal band.
Beginning with the 180-degree meridian and proceeding eastward,
the bands are numbered from 01 to 60, so that 180 E to 174 W is band 001;
174 W to 168 W is band 002; and so on.

The fourth and fifth characters designate a 6-degree wide latitudinal band.
Beginning at the south pole and proceeding northward, the bands are lettered
from AA to BK (using only letters ABCDEFGHJKLMNPQRSTUV) so that 90 S to 84 S is band AA;
84 S to 78 S is band AB; and so on.

# 3 DEGREE CELLS

    0  15  30
 30 ---------
    | 1 | 2 |
 15 ---------
    | 3 | 4 |
 0  ---------

The quadrants are numbered sequentially, from west to east,
starting with the northernmost band. Specifically, the northwest quadrant is 1;
the northeast quadrant is 2; the southwest quadrant is 3; the southeast quadrant is 4.

Each quadrant is identified by a six-character designation. (ex. D06AG3)
The first five characters comprise the 6-degree cell designation.
The sixth character is the quadrant number.

# 1 DEGREE CELLS

    0   5  10  15
 15 -------------
    | 1 | 2 | 3 |
 10 -------------
    | 4 | 5 | 6 |
 5  -------------
    | 7 | 8 | 9 |
 0  -------------

The areas are numbered sequentially, from west to east, starting with the northernmost band.
The graphical representation of a 3-degree quadrant with numbered 1-degree by 1-degree
areas resembles a telephone keypad

Each 1-degree by 1-degree area, or keypad key is identified by a seven-character
designation. The first six characters comprise the 3-degree quadrant designation.
The seventh character is the keypad key number. (ex. D06AG39)

"""
import math
import re
from typing import Optional

import shapely.geometry

from .garsgrid import GARSGridBase


class EDGARSGrid(GARSGridBase):
    """
    This object gives the polygon of the ED-GARS grid based on the ED-GARS ID
    or the resolution and lat/lon coords for the extended version using
    degrees.

    """

    LETTERS = "ABCDEFGHJKLMNPQRSTUV"
    VALID_RESOLUTIONS = (1, 3, 6)
    RE_PATTERN = re.compile(
        r"^D(?P<quadrant_6deg>\d{2}[A-B][A-HJ-NP-V])"
        r"((?P<quadrant_3deg>[1-9])"
        r"(?P<quadrant_1deg>[1-9])?)?$"
    )

    def __init__(self, gars_id: str, max_resolution: Optional[int] = None) -> None:
        """
        Parameters
        ----------
        gars_id: str
            The string representing the ED-GARS grid.
        max_resolution: int, optional
            The grid resolution in degrees (1, 3, 6).
            If not provided, it will be inferred from the `gars_id`.

        """
        super().__init__()

        max_resolution = (
            None if max_resolution is None else self.validate_resolution(max_resolution)
        )

        # truncate ED-GARS ID to match resolution
        if max_resolution == 6:
            gars_id = gars_id[:5]
        elif max_resolution == 3:
            gars_id = gars_id[:6]

        # validate ED-GARS ID
        gars_match = self.RE_PATTERN.match(gars_id)
        if not gars_match:
            raise ValueError(f'"{gars_id}" is not a valid ED-GARS grid ID.')
        self.gars_id: str = gars_id

        gars_dict = gars_match.groupdict()

        self.quadrant_1deg: str = gars_dict["quadrant_1deg"]
        self.quadrant_3deg: str = gars_dict["quadrant_3deg"]
        self.quadrant_6deg: str = gars_dict["quadrant_6deg"]

        lon_num = int(self.quadrant_6deg[:2])
        if (lon_num < 1) or (lon_num > 60):
            raise ValueError(
                f'"{gars_id}" is not a valid ED-GARS grid ID. '
                "Longitude numbers can only be between 01-60."
            )

        if self.quadrant_6deg[2] == "B" and self.quadrant_6deg[3] > "K":
            raise ValueError(
                f'"{gars_id}" is not a valid ED-GARS grid ID. '
                "Latitude letters cannot exceed BK."
            )

        # determine resolution from ED-GARS ID
        self.resolution: int = 6
        if self.quadrant_1deg:
            self.resolution = 1
        elif self.quadrant_3deg:
            self.resolution = 3

        # properties
        self._polygon: Optional[shapely.geometry.Polygon] = None

    def __repr__(self) -> str:
        return f"<ED-GARS(gars_id={self}, resolution={self.resolution})>"

    @classmethod
    def from_latlon(
        cls, latitude: float, longitude: float, resolution: int
    ) -> "EDGARSGrid":
        """Load ED-GARS grid from latitude and longitude.

        Parameters
        ----------
        latitude: float
            The latitude of the cell you want to find.
        longitude: float
            The longitude of the cell you want to find.
        resolution: int
            The grid resolution in degrees (1, 3, 6).


        Returns
        -------
        EDGARSGrid

        """
        resolution = cls.validate_resolution(resolution)

        # convert from (-180, 180) to (0, 360)
        longitude = (longitude + 180) % 360 if longitude != 180 else 360
        # convert from (-90, 90) to (0, 180)
        latitude = (latitude + 90) % 180 if latitude != 90 else 179.9999999999

        lon_idx = longitude / 6.0
        lat_idx = latitude / 6.0
        # 6 deg quadrant
        quadrant_6deg = (
            f"{int(lon_idx + 1):02d}"
            + cls.LETTERS[int(math.floor(lat_idx // 20))]
            + cls.LETTERS[int(math.floor(lat_idx % 20))]
        )

        quadrant_3deg = ""
        quadrant_1deg = ""
        # 3 deg quadrant
        if resolution < 6:
            lon_3deg_idx = math.floor((longitude % 6) / 3.0) + 1
            lat_3deg_idx = 2 - math.floor((latitude % 6) / 3.0)
            quadrant_3deg = str(int((lat_3deg_idx - 1) * 2 + lon_3deg_idx))

            # 1 deg quadrant
            if resolution < 3:
                lon_1deg_idx = math.floor(longitude % 3) + 1
                lat_1deg_idx = 3 - math.floor(latitude % 3)
                quadrant_1deg = str(int((lat_1deg_idx - 1) * 3 + lon_1deg_idx))

        gars_id = "".join(["D", quadrant_6deg, quadrant_3deg, quadrant_1deg])

        return cls(gars_id, max_resolution=resolution)

    @property
    def polygon(self) -> shapely.geometry.Polygon:
        """Generates the GARS bounding polygon.

        Returns
        -------
        :obj:`shapely.geometry.Polygon`

        """
        if self._polygon is not None:
            return self._polygon

        # CALCULATE 6 DEG DEGREES
        # get 6 DEG quadrant info
        longitude: float = ((int(self.quadrant_6deg[:2]) - 1) * 6) - 180

        # first 6 deg north/south letter, A-B
        lat_6deg_letter1 = self.quadrant_6deg[2]

        # second 6 deg north/south letter, A-V
        lat_6deg_letter2 = self.quadrant_6deg[3]

        latitude: float = (-90.0 + (self.LETTERS.index(lat_6deg_letter1) * 120.0)) + (
            self.LETTERS.index(lat_6deg_letter2) * 6.0
        )

        # CALCULATE 3 DEG DELTA
        try:
            quad_3deg = int(self.quadrant_3deg)
        except TypeError:
            quad_3deg = None

        if quad_3deg is not None:
            if quad_3deg in (1, 2):
                latitude += 3.0
            if quad_3deg in (2, 4):
                longitude += 3.0

        # CALCULATE 1 DEG DELTA
        try:
            quad_1deg = int(self.quadrant_1deg)
        except TypeError:
            quad_1deg = None

        if quad_1deg is not None:
            if quad_1deg in (2, 5, 8):
                longitude += 1.0
            elif quad_1deg in (3, 6, 9):
                longitude += 2.0

            if quad_1deg in (4, 5, 6):
                latitude += 1.0
            elif quad_1deg in (1, 2, 3):
                latitude += 2.0

        # upper right coordinate
        ul_latitude = latitude + self.resolution
        ul_longitude = longitude + self.resolution

        self._polygon = shapely.geometry.box(
            longitude, latitude, ul_longitude, ul_latitude
        )

        return self._polygon



"""
This module contains tools to generate
 a set of GARS grids (field) based on:
- the bounding area of interest
- the desired resolution.

"""
import itertools
from typing import List, Optional, Tuple, Type

import shapely

from .edgarsgrid import EDGARSGrid
from .garsgrid import GARSGrid, GARSGridBase
from .gedgarsgrid import GEDGARSGrid


class GARSField:
    """
    Given a Shapely shape object and an GARS grid resolution, return a set of
    GARSGrid objects which intersect.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, bounding_geom: shapely.geometry.base.BaseGeometry) -> None:
        """
        Parameters
        ----------
        bounding_geom: :obj:`shapely.geometry.base.BaseGeometry`
            The geometry to retrieve grids from.

        """
        self.bounding_geom = bounding_geom

        self._gars_60deg: Optional[List[GEDGARSGrid]] = None
        self._gars_30deg: Optional[List[GEDGARSGrid]] = None

        self._gars_6deg: Optional[List[EDGARSGrid]] = None
        self._gars_3deg: Optional[List[EDGARSGrid]] = None
        self._gars_1deg: Optional[List[EDGARSGrid]] = None

        self._gars_30min: Optional[List[GARSGrid]] = None
        self._gars_15min: Optional[List[GARSGrid]] = None
        self._gars_5min: Optional[List[GARSGrid]] = None
        self._gars_1min: Optional[List[GARSGrid]] = None

    def _get_bounds(self) -> Tuple[float, float, float, float]:
        """
        Retrieve the bounding coordinates of the input geometry

        Returns:
        --------
        tuple:
            min_lon, min_lat, max_lon, max_lat
        """
        min_lon, min_lat, max_lon, max_lat = self.bounding_geom.bounds

        if min_lon < -180:
            min_lon = -180
        if max_lon >= 180:
            max_lon = 179.999999999999

        return min_lon, min_lat, max_lon, max_lat

    @staticmethod
    def _get_lat_letter_range(
        gars_grid: Type[GARSGridBase],
        ll_lat1: str,
        ll_lat2: str,
        ur_lat1: str,
        ur_lat2: str,
    ) -> List[str]:
        """
        Retrieve the latitude letter range for the bounding box.
        """
        # ignoring types due to: https://github.com/python/mypy/issues/4125

        # first north/south letter,  GARSGrid (A-Q), EDGARSGRid (A-B)
        lat_letter1_range: str = gars_grid.LETTERS[  # type: ignore
            gars_grid.LETTERS.index(ll_lat1) : gars_grid.LETTERS.index(ur_lat1) + 1  # type: ignore
        ]

        # second north/south letter, A-Z
        lat_letter2_range: str = gars_grid.LETTERS[  # type: ignore
            gars_grid.LETTERS.index(ll_lat2) : gars_grid.LETTERS.index(ur_lat2) + 1  # type: ignore
        ]

        lat_letter_range: List[str] = []
        if ll_lat1 != ur_lat1:
            # part 1: using first letter1 and all from first letter2 to end of all LETTERS
            for letter2 in gars_grid.LETTERS[gars_grid.LETTERS.index(ll_lat2) :]:  # type: ignore
                lat_letter_range.append("".join([ll_lat1, letter2]))

            # part 2: from second letter1 to next to last letter2 with all LETTERS
            for letter1, letter2 in itertools.product(
                lat_letter1_range[1:-1], gars_grid.LETTERS  # type: ignore
            ):
                lat_letter_range.append("".join([letter1, letter2]))

            # part 3: using last letter1 and all from beginning of all LETTERS to last last letter2
            for letter2 in gars_grid.LETTERS[: gars_grid.LETTERS.index(ur_lat2) + 1]:  # type: ignore
                lat_letter_range.append("".join([ur_lat1, letter2]))

        else:
            for letter1, letter2 in itertools.product(
                lat_letter1_range, lat_letter2_range
            ):
                lat_letter_range.append("".join([letter1, letter2]))
        return lat_letter_range

    @property
    def gars_60deg(self) -> List[GEDGARSGrid]:
        """list: The 36deg GEDGARSGrid objects that intersect the bounding geometry."""
        if self._gars_60deg is not None:
            return self._gars_60deg

        # handle multi geometry
        if self.bounding_geom.type.startswith("Multi"):
            self._gars_60deg = []
            for bounding_geom in self.bounding_geom.geoms:
                self._gars_60deg += GARSField(bounding_geom).gars_60deg
            # make sure unique
            self._gars_60deg = sorted(set(self._gars_60deg))
            return self._gars_60deg

        # generate the list of 36 degree GED-GARS grids
        self._gars_60deg = []
        for lon_num_60deg, lat_60deg_letter in itertools.product(
            range(1, 7), GEDGARSGrid.LETTERS
        ):
            g60deg = GEDGARSGrid("".join(["GD", str(lon_num_60deg), lat_60deg_letter]))
            if g60deg.polygon.intersects(self.bounding_geom):
                self._gars_60deg.append(g60deg)

        return self._gars_60deg

    @property
    def gars_30deg(self) -> List[GEDGARSGrid]:
        """list: The 30deg GEDGARSGrid objects that intersect the bounding geometry."""
        if self._gars_30deg is not None:
            return self._gars_30deg

        # handle multi geometry
        if self.bounding_geom.type.startswith("Multi"):
            self._gars_30deg = []
            for bounding_geom in self.bounding_geom.geoms:
                self._gars_30deg += GARSField(bounding_geom).gars_30deg
            # make sure unique
            self._gars_30deg = sorted(set(self._gars_30deg))
            return self._gars_30deg

        # generate the list of 3 degree ED-GARS grids
        self._gars_30deg = []
        for gars_60deg, quadrant_30deg in itertools.product(
            self.gars_60deg, range(1, 5)
        ):
            g30deg = GEDGARSGrid("".join([str(gars_60deg), str(quadrant_30deg)]))
            if g30deg.polygon.intersects(self.bounding_geom):
                self._gars_30deg.append(g30deg)
        return self._gars_30deg

    def _get_6deg_ranges(self) -> Tuple[List[str], List[str]]:
        """Retrieves the latitude numeric range and longitude letter range
        within the bounds.
        """
        min_lon, min_lat, max_lon, max_lat = self._get_bounds()

        ll_gars = str(EDGARSGrid.from_latlon(min_lat, min_lon, resolution=6))
        ur_gars = str(EDGARSGrid.from_latlon(max_lat, max_lon, resolution=6))

        lon_num_range = [
            f"{lon_num:02d}"
            for lon_num in range(int(ll_gars[1:3]), int(ur_gars[1:3]) + 1)
        ]

        lat_letter_range = self._get_lat_letter_range(
            EDGARSGrid, ll_gars[3], ll_gars[4], ur_gars[3], ur_gars[4]
        )

        return lon_num_range, lat_letter_range

    @property
    def gars_6deg(self) -> List[EDGARSGrid]:
        """list: The 6deg EDGARSGrid objects that intersect the bounding geometry."""
        if self._gars_6deg is not None:
            return self._gars_6deg

        # handle multi geometry
        if self.bounding_geom.type.startswith("Multi"):
            self._gars_6deg = []
            for bounding_geom in self.bounding_geom.geoms:
                self._gars_6deg += GARSField(bounding_geom).gars_6deg
            # make sure unique
            self._gars_6deg = sorted(set(self._gars_6deg))
            return self._gars_6deg

        # generate the list of 6 degree ED-GARS grids
        self._gars_6deg = []
        for lon_num_6deg, lat_6deg_letter in itertools.product(
            *self._get_6deg_ranges()
        ):
            try:
                g6deg = EDGARSGrid("".join(["D", lon_num_6deg, lat_6deg_letter]))
            except ValueError:
                continue
            if g6deg.polygon.intersects(self.bounding_geom):
                self._gars_6deg.append(g6deg)

        return self._gars_6deg

    @property
    def gars_3deg(self) -> List[EDGARSGrid]:
        """list: The 3deg EDGARSGrid objects that intersect the bounding geometry."""
        if self._gars_3deg is not None:
            return self._gars_3deg

        # handle multi geometry
        if self.bounding_geom.type.startswith("Multi"):
            self._gars_3deg = []
            for bounding_geom in self.bounding_geom.geoms:
                self._gars_3deg += GARSField(bounding_geom).gars_3deg
            # make sure unique
            self._gars_3deg = sorted(set(self._gars_3deg))
            return self._gars_3deg

        # generate the list of 3 degree ED-GARS grids
        self._gars_3deg = []
        for gars_6deg, quadrant_3deg in itertools.product(self.gars_6deg, range(1, 5)):
            g3deg = EDGARSGrid("".join([str(gars_6deg), str(quadrant_3deg)]))
            if g3deg.polygon.intersects(self.bounding_geom):
                self._gars_3deg.append(g3deg)
        return self._gars_3deg

    @property
    def gars_1deg(self) -> List[EDGARSGrid]:
        """list: The 1deg EDGARSGrid objects that intersect the bounding geometry."""
        if self._gars_1deg is not None:
            return self._gars_1deg

        # handle multi geometry
        if self.bounding_geom.type.startswith("Multi"):
            self._gars_1deg = []
            for bounding_geom in self.bounding_geom.geoms:
                self._gars_1deg += GARSField(bounding_geom).gars_1deg
            # make sure unique
            self._gars_1deg = sorted(set(self._gars_1deg))
            return self._gars_1deg

        # generate the list of 1 degree ED-GARS grids
        self._gars_1deg = []
        for gars_3deg, quadrant_1deg in itertools.product(self.gars_3deg, range(1, 10)):
            g1deg = EDGARSGrid("".join([str(gars_3deg), str(quadrant_1deg)]))
            if g1deg.polygon.intersects(self.bounding_geom):
                self._gars_1deg.append(g1deg)
        return self._gars_1deg

    def _get_30min_ranges(self) -> Tuple[List[str], List[str]]:
        """Retrieves the latitude numeric range and longitude letter range
        within the bounds.
        """
        min_lon, min_lat, max_lon, max_lat = self._get_bounds()

        ll_gars = str(GARSGrid.from_latlon(min_lat, min_lon, resolution=30))
        ur_gars = str(GARSGrid.from_latlon(max_lat, max_lon, resolution=30))

        lon_num_range = [
            f"{lon_num:03d}"
            for lon_num in range(int(ll_gars[:3]), int(ur_gars[:3]) + 1)
        ]

        lat_letter_range = self._get_lat_letter_range(
            GARSGrid, ll_gars[3], ll_gars[4], ur_gars[3], ur_gars[4]
        )

        return lon_num_range, lat_letter_range

    @property
    def gars_30min(self) -> List[GARSGrid]:
        """list: The 30min GARSGrid objects that intersect the bounding geometry."""
        if self._gars_30min is not None:
            return self._gars_30min

        # handle multi geometry
        if self.bounding_geom.type.startswith("Multi"):
            self._gars_30min = []
            for bounding_geom in self.bounding_geom.geoms:
                self._gars_30min += GARSField(bounding_geom).gars_30min
            # make sure unique
            self._gars_30min = sorted(set(self._gars_30min))
            return self._gars_30min

        # generate the list of 30 minute GARS grids
        self._gars_30min = []
        for lon_num_30min, lat_30min_letter in itertools.product(
            *self._get_30min_ranges()
        ):
            try:
                g30min = GARSGrid("".join([lon_num_30min, lat_30min_letter]))
            except ValueError:
                continue
            if g30min.polygon.intersects(self.bounding_geom):
                self._gars_30min.append(g30min)

        return self._gars_30min

    @property
    def gars_15min(self) -> List[GARSGrid]:
        """list: The 15min GARSGrid objects that intersect the bounding geometry."""
        if self._gars_15min is not None:
            return self._gars_15min

        # handle multi geometry
        if self.bounding_geom.type.startswith("Multi"):
            self._gars_15min = []
            for bounding_geom in self.bounding_geom.geoms:
                self._gars_15min += GARSField(bounding_geom).gars_15min
            # make sure unique
            self._gars_15min = sorted(set(self._gars_15min))
            return self._gars_15min

        # generate the list of 15 minute GARS grids
        self._gars_15min = []
        for gars_30min, quadrant_15min in itertools.product(
            self.gars_30min, range(1, 5)
        ):
            g15min = GARSGrid("".join([str(gars_30min), str(quadrant_15min)]))
            if g15min.polygon.intersects(self.bounding_geom):
                self._gars_15min.append(g15min)
        return self._gars_15min

    @property
    def gars_5min(self) -> List[GARSGrid]:
        """list: The 5min GARSGrid objects that intersect the bounding geometry."""
        if self._gars_5min is not None:
            return self._gars_5min

        # handle multi geometry
        if self.bounding_geom.type.startswith("Multi"):
            self._gars_5min = []
            for bounding_geom in self.bounding_geom.geoms:
                self._gars_5min += GARSField(bounding_geom).gars_5min
            # make sure unique
            self._gars_5min = sorted(set(self._gars_5min))
            return self._gars_5min

        # generate the list of 5 minute GARS grids
        self._gars_5min = []
        for gars_15min, quadrant_5min in itertools.product(
            self.gars_15min, range(1, 10)
        ):
            g5min = GARSGrid("".join([str(gars_15min), str(quadrant_5min)]))
            if g5min.polygon.intersects(self.bounding_geom):
                self._gars_5min.append(g5min)
        return self._gars_5min

    @property
    def gars_1min(self) -> List[GARSGrid]:
        """list: The 1min GARSGrid objects that intersect the bounding geometry."""
        if self._gars_1min is not None:
            return self._gars_1min

        # handle multi geometry
        if self.bounding_geom.type.startswith("Multi"):
            self._gars_1min = []
            for bounding_geom in self.bounding_geom.geoms:
                self._gars_1min += GARSField(bounding_geom).gars_1min
            # make sure unique
            self._gars_1min = sorted(set(self._gars_1min))
            return self._gars_1min

        # generate the list of 1 minute GARS grids
        self._gars_1min = []
        for gars_5min, quadrant_1min in itertools.product(self.gars_5min, range(1, 26)):
            g1min = GARSGrid("".join([str(gars_5min), f"{quadrant_1min:02d}"]))
            if g1min.polygon.intersects(self.bounding_geom):
                self._gars_1min.append(g1min)
        return self._gars_1min



import pytest

from gars_field import EDGARSGrid


@pytest.mark.parametrize(
    "gars_id,long_gars_id,expected_resolution",
    [
        ("D20BJ26", "D20BJ26", 1),
        ("D20BJ2", "D20BJ26", 3),
        ("D20BJ", "D20BJ2", 6),
        ("D20BJ", "D20BJ26", 6),
    ],
)
def test_garsd(gars_id, long_gars_id, expected_resolution):
    # check resolution calc
    gg = EDGARSGrid(gars_id)
    assert str(gg) == gars_id
    assert gg.resolution == expected_resolution

    # check resolution forcing
    ggf = EDGARSGrid(long_gars_id, max_resolution=expected_resolution)
    assert str(ggf) == gars_id
    assert ggf.resolution == expected_resolution


@pytest.mark.parametrize(
    "lat,lon,resolution,expected_id",
    [
        (-89.55, -179.57, 1, "D01AA37"),
        (-89.55, -179.57, 3, "D01AA3"),
        (-89.55, -179.57, 6, "D01AA"),
        (-90, -179, 1, "D01AA38"),
        (89, -180, 3, "D01BK1"),
        (-90, 179.55, 1, "D60AA49"),
        (-90, 179, 1, "D60AA49"),
        (0.005, 179.9, 1, "D60AR49"),
        (0.303, 179.9, 1, "D60AR49"),
    ],
)
def test_garsd_from_latlon(lat, lon, resolution, expected_id):
    assert str(EDGARSGrid.from_latlon(lat, lon, resolution=resolution)) == expected_id


@pytest.mark.parametrize(
    "gars_id, expected_polygon",
    [
        ("D20BJ26", "POLYGON ((-60 82, -60 83, -61 83, -61 82, -60 82))"),
        ("D20BJ2", "POLYGON ((-60 81, -60 84, -63 84, -63 81, -60 81))"),
        ("D20BJ", "POLYGON ((-60 78, -60 84, -66 84, -66 78, -60 78))"),
    ],
)
def test_garsd_polygon(gars_id, expected_polygon):
    assert str(EDGARSGrid(gars_id).polygon) == expected_polygon


@pytest.mark.parametrize(
    "invalid_gars_id", ["720HN96", "D20BN26", "D80AB26", "D20BJ999"]
)
def test_garsd_to_polygon__invalid_id(invalid_gars_id):
    with pytest.raises(ValueError):
        EDGARSGrid(invalid_gars_id)


@pytest.mark.parametrize(
    "gars_id, expected_utm_epsg",
    [
        ("D60AG26", "EPSG:32760"),
        ("D21AC2", "EPSG:32721"),
        ("D04BA", "EPSG:32604"),
        ("D20AB", "EPSG:32761"),
        ("D45BK", "EPSG:32661"),
    ],
)
def test_garsd_utm_epsg(gars_id, expected_utm_epsg):
    assert str(EDGARSGrid(gars_id).utm_epsg) == expected_utm_epsg



"""
GARS Field is a toolkit for generating Global Area Reference grids
"""
__version__ = "0.3.2"

from .edgarsgrid import EDGARSGrid  # noqa: F401
from .field import GARSField  # noqa: F401
from .garsgrid import GARSGrid  # noqa: F401
from .gedgarsgrid import GEDGARSGrid  # noqa: F401



"""
This utility is designed to generate a GARS grid.

Code Based on:
- https://github.com/mil-oss/GARSutils
- https://github.com/Moustikitos/gryd/blob/c79edde94f19d46e3b3532ae14eb351e91d55322/Gryd/geodesy.py

Design Information From:
- http://earth-info.nga.mil/GandG/coordsys/grids/gars.html

# 30 MINUTE CELLS
GARS divides the surface of the earth into 30-minute by 30-minute cells.
Each cell is identified by a five-character designation. (ex. 006AG)

The first three characters designate a 30-minute wide longitudinal band.
Beginning with the 180-degree meridian and proceeding eastward,
the bands are numbered from 001 to 720, so that 180 E to 179 30 W is band 001;
179 30 W to 179 00 W is band 002; and so on.

The fourth and fifth characters designate a 30-minute wide latitudinal band.
Beginning at the south pole and proceeding northward, the bands are lettered
from AA to QZ (omitting I and O) so that 90 00 S to 89 30 S is band AA;
89 30 S to 89 00 S is band AB; and so on.

# 15 MINUTE CELLS

    0  15  30
 30 ---------
    | 1 | 2 |
 15 ---------
    | 3 | 4 |
 0  ---------

The quadrants are numbered sequentially, from west to east,
starting with the northernmost band. Specifically, the northwest quadrant is 1;
the northeast quadrant is 2; the southwest quadrant is 3; the southeast quadrant is 4.

Each quadrant is identified by a six-character designation. (ex. 006AG3)
The first five characters comprise the 30-minute cell designation.
The sixth character is the quadrant number.

# 5 MINUTE CELLS

    0   5  10  15
 15 -------------
    | 1 | 2 | 3 |
 10 -------------
    | 4 | 5 | 6 |
 5  -------------
    | 7 | 8 | 9 |
 0  -------------

The areas are numbered sequentially, from west to east, starting with the northernmost band.
The graphical representation of a 15-minute quadrant with numbered 5-minute by 5-minute
areas resembles a telephone keypad

Each 5-minute by 5-minute area, or keypad key is identified by a seven-character
designation. The first six characters comprise the 15-minute quadrant designation.
The seventh character is the keypad key number. (ex.006AG39)


# 1 MINUTE CELLS (CUSTOM EXTENSION)

    0   1   2   3   4   5
 5  ---------------------
    | 01| 02| 03| 04| 05|
 4  ---------------------
    | 06| 07| 08| 09| 10|
 3  ---------------------
    | 11| 12| 13| 14| 15|
 2  ---------------------
    | 16| 17| 18| 19| 20|
 1  ---------------------
    | 21| 22| 23| 24| 25|
 0  ---------------------

The areas are numbered sequentially, from west to east, starting with the northernmost band.
The graphical representation of a 5-minute quadrant with numbered 1-minute by 1-minute
areas resembles a telephone keypad

Each 1-minute by 1-minute area, or keypad key is identified by a seven-character
designation. The first seven characters comprise the 5-minute quadrant designation.
The eighth and ninth characters are the keypad key number. (ex.006AG3901)

"""
import math
import re
from abc import abstractmethod
from typing import Optional, Tuple

import shapely.geometry
from pyproj.aoi import AreaOfInterest
from pyproj.database import (  # type: ignore  # pylint: disable=no-name-in-module
    query_utm_crs_info,
)


class GARSGridBase:
    """
    Base class for all GARS Grid classes.
    """

    def __init__(self):
        self._utm_epsg: Optional[str] = None
        self.gars_id: str = ""

    @property
    @abstractmethod
    def polygon(self) -> shapely.geometry.Polygon:
        """
        Returns the polygon boundary for the GARS tile.

        Returns
        -------
        :obj:`shapely.geometry.Polygon`
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def VALID_RESOLUTIONS(self) -> Tuple[int, ...]:  # pylint: disable=invalid-name
        """
        List of supported resolutions.

        Returns
        -------
        tuple
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def LETTERS(self) -> str:  # pylint: disable=invalid-name
        """
        String of letters for the grid system

        Returns
        -------
        str
        """
        raise NotImplementedError

    @property
    def utm_epsg(self) -> str:
        """This returns the UTM zone EPSG code for this GARS grid.

        Returns
        -------
        str:
            UTM zone EPSG code
        """
        if self._utm_epsg is not None:
            return self._utm_epsg

        latitude = self.polygon.centroid.y
        if latitude <= -80:
            # UPS South
            self._utm_epsg = "EPSG:32761"
        elif latitude >= 84:
            # UPS North
            self._utm_epsg = "EPSG:32661"
        else:
            longitude = self.polygon.centroid.x
            utm_crs_list = query_utm_crs_info(
                datum_name="WGS 84",
                area_of_interest=AreaOfInterest(
                    west_lon_degree=longitude,
                    south_lat_degree=latitude,
                    east_lon_degree=longitude,
                    north_lat_degree=latitude,
                ),
            )
            self._utm_epsg = f"EPSG:{utm_crs_list[0].code}"
        return self._utm_epsg

    def __str__(self) -> str:
        return self.gars_id

    def __eq__(self, other) -> bool:
        return str(self) == str(other)

    def __hash__(self) -> int:
        return hash(self.gars_id)

    def __lt__(self, other) -> bool:
        return self.gars_id < other.gars_id

    @classmethod
    def validate_resolution(cls, resolution: int) -> int:
        """Validate the resolution input.

        Returns
        -------
        int: Valid resolution.

        Raises
        ------
        ValueError

        """
        if resolution not in cls.VALID_RESOLUTIONS:  # type: ignore
            raise ValueError(
                f"Invalid resolution {resolution}. "
                f"Only {cls.VALID_RESOLUTIONS} are allowed."
            )
        return int(resolution)


class GARSGrid(GARSGridBase):
    """
    This object gives the polygon of the GARS grid based on the GARS ID
    or the resolution and lat/lon coords.
    """

    LETTERS: str = "ABCDEFGHJKLMNPQRSTUVWXYZ"
    VALID_RESOLUTIONS: Tuple[int, int, int, int] = (1, 5, 15, 30)
    RE_PATTERN = re.compile(
        r"^(?P<quadrant_30min>\d{3}[A-HJ-NP-Q][A-HJ-NP-Z])"
        r"((?P<quadrant_15min>[1-4])"
        r"((?P<quadrant_5min>[1-9])"
        r"(?P<quadrant_1min>\d{2})?)?)?$"
    )

    def __init__(self, gars_id: str, max_resolution: Optional[int] = None):
        """
        Parameters
        ----------
        gars_id: str
            The string representing the GARS grid.
        max_resolution: int, optional
            The grid resolution in seconds (1, 5, 15, 30).
            If not provided, it will be inferred from the `gars_id`.

        """
        super().__init__()

        max_resolution = (
            None if max_resolution is None else self.validate_resolution(max_resolution)
        )

        # truncate GARS ID to match resolution
        if max_resolution == 30:
            gars_id = gars_id[:5]
        elif max_resolution == 15:
            gars_id = gars_id[:6]
        elif max_resolution == 5:
            gars_id = gars_id[:7]

        # validate GARS ID
        gars_match = self.RE_PATTERN.match(gars_id)
        if not gars_match:
            raise ValueError(f'"{gars_id}" is not a valid GARS grid ID.')
        self.gars_id = gars_id

        gars_dict = gars_match.groupdict()

        self.quadrant_1min = gars_dict["quadrant_1min"]
        self.quadrant_5min = gars_dict["quadrant_5min"]
        self.quadrant_15min = gars_dict["quadrant_15min"]
        self.quadrant_30min = gars_dict["quadrant_30min"]

        lon_num = int(self.quadrant_30min[:3])
        if (lon_num < 1) or (lon_num > 720):
            raise ValueError(
                f'"{gars_id}" is not a valid GARS grid ID. '
                "Longitude numbers can only be between 001-720."
            )

        # determine resolution from GARS ID
        self.resolution = 30
        if self.quadrant_1min:
            if int(self.quadrant_1min) < 1 or int(self.quadrant_1min) > 25:
                raise ValueError(
                    f'"{gars_id}" is not a valid GARS grid ID. '
                    "1 min quadrant number can only be between 01-25."
                )

            self.resolution = 1
        elif self.quadrant_5min:
            self.resolution = 5
        elif self.quadrant_15min:
            self.resolution = 15

        # properties
        self._polygon = None

    def __repr__(self) -> str:
        return f"<GARS(gars_id={self}, resolution={self.resolution})>"

    @classmethod
    def from_latlon(
        cls, latitude: float, longitude: float, resolution: int
    ) -> "GARSGrid":
        # pylint: disable=too-many-locals
        """Load GARS grid from latitude and longitude.

        Parameters
        ----------
        latitude: float
            The latitude of the cell you want to find.
        longitude: float
            The longitude of the cell you want to find.
        resolution: int
            The grid resolution in minutes (5, 15, 30).


        Returns
        -------
        GARSGrid

        """
        resolution = cls.validate_resolution(resolution)

        # convert from (-180, 180) to (0, 360)
        longitude = (longitude + 180) % 360 if longitude != 180 else 360
        # convert from (-90, 90) to (0, 180)
        latitude = (latitude + 90) % 180 if latitude != 90 else 179.9999999999

        lon_idx = longitude * 2.0
        lat_idx = latitude * 2.0

        # 30 minute quadrant
        quadrant_30min = (
            f"{int(lon_idx + 1):03d}"
            + cls.LETTERS[int(math.floor(lat_idx // 24))]
            + cls.LETTERS[int(math.floor(lat_idx % 24))]
        )

        quadrant_15min = ""
        quadrant_5min = ""
        quadrant_1min = ""
        # 15 minute quadrant
        if resolution < 30:

            def index_from_degrees(
                num_degrees: float, inverse: bool = False
            ) -> Tuple[int, int, int]:
                minutes = (num_degrees - math.floor(num_degrees)) * 60
                minutes_30 = minutes % 30
                minutes_15 = minutes % 15
                minutes_5 = minutes % 5
                idx_15 = math.floor(minutes_30 / 15.0) + 1
                idx_5 = math.floor(minutes_15 / 5.0) + 1
                idx_1 = math.floor(minutes_5) + 1
                if inverse:
                    idx_15 = 3 - idx_15
                    idx_5 = 4 - idx_5
                    idx_1 = 6 - idx_1
                return idx_15, idx_5, idx_1

            lon_15min_idx, lon_5min_idx, lon_1min_idx = index_from_degrees(longitude)
            lat_15min_idx, lat_5min_idx, lat_1min_idx = index_from_degrees(
                latitude, inverse=True
            )
            quadrant_15min = str(int((lat_15min_idx - 1) * 2 + lon_15min_idx))

            # 5 minute quadrant
            if resolution < 15:
                quadrant_5min = str(int((lat_5min_idx - 1) * 3 + lon_5min_idx))

            # 1 minute quadrant
            if resolution < 5:
                quadrant_1min = f"{int((lat_1min_idx - 1) * 5 + lon_1min_idx):02d}"

        gars_id = "".join(
            [quadrant_30min, quadrant_15min, quadrant_5min, quadrant_1min]
        )

        return cls(gars_id, max_resolution=resolution)

    def _15_minute_delta(self) -> Tuple[float, float]:
        """
        Calculate 15 minute delta

        Returns
        -------
        tuple:
            lon_minutes, lat_minutes
        """
        lat_minutes = 0.0
        lon_minutes = 0.0

        try:
            quad_15min = int(self.quadrant_15min)
        except TypeError:
            return lon_minutes, lat_minutes

        if quad_15min in (1, 2):
            lat_minutes = 15.0
        if quad_15min in (2, 4):
            lon_minutes = 15.0
        return lon_minutes, lat_minutes

    def _5_minute_delta(self) -> Tuple[float, float]:
        """
        Calculate 5 minute delta

        Returns
        -------
        tuple:
            lon_minutes, lat_minutes
        """
        lat_minutes = 0.0
        lon_minutes = 0.0

        try:
            quad_5min = int(self.quadrant_5min)
        except TypeError:
            return lon_minutes, lat_minutes

        if quad_5min in (2, 5, 8):
            lon_minutes = 5.0
        elif quad_5min in (3, 6, 9):
            lon_minutes = 10.0

        if quad_5min in (4, 5, 6):
            lat_minutes = 5.0
        elif quad_5min in (1, 2, 3):
            lat_minutes = 10.0
        return lon_minutes, lat_minutes

    def _1_minute_delta(self) -> Tuple[float, float]:
        """
        Calculate 1 minute delta

        Returns
        -------
        tuple:
            lon_minutes, lat_minutes
        """
        lat_minutes = 0.0
        lon_minutes = 0.0

        try:
            quad_1min = int(self.quadrant_1min)
        except TypeError:
            return lon_minutes, lat_minutes

        if quad_1min in (2, 7, 12, 17, 22):
            lon_minutes = 1.0
        elif quad_1min in (3, 8, 13, 18, 23):
            lon_minutes = 2.0
        elif quad_1min in (4, 9, 14, 19, 24):
            lon_minutes = 3.0
        elif quad_1min in (5, 10, 15, 20, 25):
            lon_minutes = 4.0

        if quad_1min <= 5:
            lat_minutes = 4.0
        elif 6 <= quad_1min <= 10:
            lat_minutes = 3.0
        elif 11 <= quad_1min <= 15:
            lat_minutes = 2.0
        elif 16 <= quad_1min <= 20:
            lat_minutes = 1.0
        return lon_minutes, lat_minutes

    @property
    def polygon(self) -> shapely.geometry.Polygon:
        """Generates the GARS bounding polygon.

        Returns
        -------
        :obj:`shapely.geometry.Polygon`

        """
        if self._polygon is not None:
            return self._polygon

        # CALCULATE 30 MIN DEGREES
        # get 30 minute quadrant info
        longitude = ((int(self.quadrant_30min[:3]) - 1) / 2.0) - 180

        # first 30 minute north/south letter, A-Q
        lat_30min_letter1 = self.quadrant_30min[3]

        # second 30 minute north/south letter, A-Z
        lat_30min_letter2 = self.quadrant_30min[4]

        latitude = (-90.0 + (self.LETTERS.index(lat_30min_letter1) * 12.0)) + (
            self.LETTERS.index(lat_30min_letter2) / 2.0
        )

        lon_minutes_15, lat_minutes_15 = self._15_minute_delta()
        lon_minutes_5, lat_minutes_5 = self._5_minute_delta()
        lon_minutes_1, lat_minutes_1 = self._1_minute_delta()

        # lower left coordinate
        ll_latitude = latitude + (lat_minutes_15 + lat_minutes_5 + lat_minutes_1) / 60.0
        ll_longitude = (
            longitude + (lon_minutes_15 + lon_minutes_5 + lon_minutes_1) / 60.0
        )

        # upper right coordinate
        ul_latitude = ll_latitude + self.resolution / 60.0
        ul_longitude = ll_longitude + self.resolution / 60.0

        self._polygon = shapely.geometry.box(
            ll_longitude, ll_latitude, ul_longitude, ul_latitude
        )

        return self._polygon



import pytest

from gars_field import GEDGARSGrid


@pytest.mark.parametrize(
    "gars_id,long_gars_id,expected_resolution",
    [("GD2B2", "GD2B2", 30), ("GD2B", "GD2B2", 60), ("GD2B", "GD2B", 60)],
)
def test_garsgd(gars_id, long_gars_id, expected_resolution):
    # check resolution calc
    gg = GEDGARSGrid(gars_id)
    assert str(gg) == gars_id
    assert gg.resolution == expected_resolution

    # check resolution forcing
    ggf = GEDGARSGrid(long_gars_id, max_resolution=expected_resolution)
    assert str(ggf) == gars_id
    assert ggf.resolution == expected_resolution


@pytest.mark.parametrize(
    "lat,lon,resolution,expected_id",
    [
        (-89.55, -179.57, 30, "GD1A3"),
        (-89.55, -179.57, 60, "GD1A"),
        (-90, -179, 30, "GD1A3"),
        (89, -180, 60, "GD1C"),
        (-90, 179.55, 30, "GD6A4"),
        (-90, 179, 60, "GD6A"),
        (0.005, 179.9, 30, "GD6B2"),
        (0.303, 179.9, 30, "GD6B2"),
    ],
)
def test_garsd_from_latlon(lat, lon, resolution, expected_id):
    assert str(GEDGARSGrid.from_latlon(lat, lon, resolution=resolution)) == expected_id


@pytest.mark.parametrize(
    "gars_id, expected_polygon",
    [
        ("GD2B2", "POLYGON ((-60 0, -60 30, -90 30, -90 0, -60 0))"),
        ("GD2B", "POLYGON ((-60 -30, -60 30, -120 30, -120 -30, -60 -30))"),
    ],
)
def test_garsd_polygon(gars_id, expected_polygon):
    assert str(GEDGARSGrid(gars_id).polygon) == expected_polygon


@pytest.mark.parametrize(
    "invalid_gars_id", ["720HN96", "D20BN26", "GD02F2", "GD8A6", "GD2B999"]
)
def test_garsd_to_polygon__invalid_id(invalid_gars_id):
    with pytest.raises(ValueError):
        GEDGARSGrid(invalid_gars_id)


def test_garsd_utm_epsg():
    assert GEDGARSGrid("GD2B").utm_epsg is None
